from typing import Dict, Optional
from google.cloud import speech
from google.cloud.speech_v1.types import RecognitionConfig, RecognitionAudio

from . import (
    LibraryContext,
    keyword,
)

ENCODING = {
    "AMR": RecognitionConfig.AudioEncoding.AMR,
    "AMR_WB": RecognitionConfig.AudioEncoding.AMR_WB,
    "FLAC": RecognitionConfig.AudioEncoding.FLAC,
    "LINEAR16": RecognitionConfig.AudioEncoding.LINEAR16,
    "MULAW": RecognitionConfig.AudioEncoding.MULAW,
    "OGG": RecognitionConfig.AudioEncoding.OGG_OPUS,
    "SPEEX": RecognitionConfig.AudioEncoding.SPEEX_WITH_HEADER_BYTE,  # noqa: E501 # pylint: disable=C0301
    "UNSPECIFIED": RecognitionConfig.AudioEncoding.ENCODING_UNSPECIFIED,  # noqa: E501 # pylint: disable=C0301
}


class SpeechToTextKeywords(LibraryContext):
    """Class for Google Cloud Speech-To-Text API

    Possible input audio encodings:

    - 'AMR'
    - 'AMR_WB'
    - 'ENCODING_UNSPECIFIED'
    - 'FLAC'
    - 'LINEAR16'
    - 'MULAW'
    - 'OGG_OPUS'
    - 'SPEEX_WITH_HEADER_BYTE'

    Link to `Speech To Text PyPI`_ page.

    .. _Speech To Text PyPI: https://pypi.org/project/google-cloud-speech/
    """

    def __init__(self, ctx):
        super().__init__(ctx)
        self.service = None

    @keyword(tags=["init", "speech to text"])
    def init_speech_to_text(
        self,
        service_account: str = None,
        use_robocorp_vault: Optional[bool] = None,
        token_file: str = None,
    ) -> None:
        """Initialize Google Cloud Speech to Text client

        :param service_account: file path to service account file
        :param use_robocorp_vault: use credentials in `Robocorp Vault`
        :param token_file: file path to token file
        """
        self.service = self.init_service_with_object(
            speech.SpeechClient, service_account, use_robocorp_vault, token_file
        )

    @keyword(tags=["speech to text"])
    def recognize_text_from_audio(
        self,
        audio_file: str = None,
        audio_uri: str = None,
        encoding: str = None,
        language_code: str = "en_US",
        audio_channel_count: int = 2,
        sample_rate: int = None,
    ) -> Dict:
        """Recognize text in the audio file

        :param audio_file: local audio file path
        :param audio_uri: Google Cloud Storage URI
        :param encoding: audio file encoding
        :param language_code: language in the audio
        :param audio_channel_count: number of audio channel
        :param sample_rate: rate in hertz, for example 16000
        :return: recognized texts

        **Examples**

        **Robot Framework**

        .. code-block:: robotframework

            ${result}=  Recognize Text From Audio   audio_file=${CURDIR}${/}test.mp3
        """
        audio = self.set_audio_type(audio_file, audio_uri)
        parameters = {"use_enhanced": True}
        # audio_encoding = ENCODING["UNSPECIFIED"]
        if encoding and encoding.upper() in ENCODING.keys():
            parameters["encoding"] = ENCODING[encoding.upper()]
        if sample_rate:
            parameters["sample_rate_hertz"] = sample_rate
        if language_code:
            parameters["language_code"] = language_code
        if audio_channel_count:
            parameters["audio_channel_count"] = audio_channel_count
        config = RecognitionConfig(**parameters)  # pylint: disable=E1101
        rec = self.service.recognize(config=config, audio=audio)
        return rec.results

    def set_audio_type(self, audio_file, audio_uri):
        # flac or wav, does not require encoding type
        if audio_file:
            with open(audio_file, "rb") as f:
                content = f.read()
                return RecognitionAudio(content=content)  # pylint: disable=E1101
        elif audio_uri:
            return RecognitionAudio(uri=audio_uri)  # pylint: disable=E1101
        else:
            raise KeyError("'audio_file' or 'audio_uri' is required")



from typing import Dict, Optional
from google.cloud import vision

from . import (
    LibraryContext,
    keyword,
)


class VisionKeywords(LibraryContext):
    """Keywords for Google Vision operations"""

    def __init__(self, ctx):
        super().__init__(ctx)
        self.service = None

    @keyword(tags=["init", "vision"])
    def init_vision(
        self,
        service_account: str = None,
        use_robocorp_vault: Optional[bool] = None,
        token_file: str = None,
    ) -> None:
        """Initialize Google Cloud Vision client

        :param service_account: file path to service account file
        :param use_robocorp_vault: use credentials in `Robocorp Vault`
        :param token_file: file path to token file
        """
        self.service = self.init_service_with_object(
            vision.ImageAnnotatorClient,
            service_account,
            use_robocorp_vault,
            token_file,
        )

    def set_image_type(self, image_file: str = None, image_uri: str = None):
        if image_file:
            with open(image_file, "rb") as f:
                content = f.read()
                return {"image": {"content": content}}
        elif image_uri:
            return {"image": {"source": {"image_uri": image_uri}}}
        else:
            raise KeyError("'image_file' or 'image_uri' is required")

    @keyword(tags=["vision"])
    def detect_labels(
        self, image_file: str = None, image_uri: str = None, json_file: str = None
    ) -> Dict:
        """Detect labels in the image

        :param image_file: source image file path
        :param image_uri: source image uri
        :param json_file: json target to save result
        :return: detection response

        **Examples**

        **Robot Framework**

        .. code-block:: robotframework

            ${result}=   Detect Labels   image_file=${CURDIR}${/}test.png
            ...  json_file=${CURDIR}${/}result.json
        """
        parameters = self.set_image_type(image_file, image_uri)
        response = self.service.label_detection(**parameters)
        self.write_json(json_file, response)
        return response

    @keyword(tags=["vision"])
    def detect_text(
        self, image_file: str = None, image_uri: str = None, json_file: str = None
    ) -> Dict:
        """Detect text in the image

        :param image_file: source image file path
        :param image_uri: Google Cloud Storage URI
        :param json_file: json target to save result
        :return: detection response

        **Examples**

        **Robot Framework**

        .. code-block:: robotframework

            ${result}=   Detect Text   image_file=${CURDIR}${/}test.png
            ...  json_file=${CURDIR}${/}result.json
        """
        parameters = self.set_image_type(image_file, image_uri)
        response = self.service.text_detection(**parameters)
        self.write_json(json_file, response)
        return response

    @keyword(tags=["vision"])
    def detect_document(
        self, image_file: str = None, image_uri: str = None, json_file: str = None
    ) -> Dict:
        """Detect document

        :param image_file: source image file path
        :param image_uri: Google Cloud Storage URI
        :param json_file: json target to save result
        :return: detection response

        **Examples**

        **Robot Framework**

        .. code-block:: robotframework

            ${result}=   Detect Document   image_file=${CURDIR}${/}test.png
            ...  json_file=${CURDIR}${/}result.json
        """
        parameters = self.set_image_type(image_file, image_uri)
        response = self.service.document_text_detection(**parameters)
        self.write_json(json_file, response)
        return response

    @keyword(tags=["vision"])
    def annotate_image(
        self, image_file: str, image_uri: str, json_file: str = None
    ) -> Dict:
        """Annotate image

        :param image_file: source image file path
        :param image_uri: Google Cloud Storage URI
        :param json_file: json target to save result
        :return: detection response

        **Examples**

        **Robot Framework**

        .. code-block:: robotframework

            ${result}=   Annotate Image   image_file=${CURDIR}${/}test.png
            ...  json_file=${CURDIR}${/}result.json
        """
        parameters = self.set_image_type(image_file, image_uri)
        response = self.service.annotate_image(**parameters)
        self.write_json(json_file, response)
        return response

    @keyword(tags=["vision"])
    def face_detection(
        self, image_file: str = None, image_uri: str = None, json_file: str = None
    ) -> Dict:
        """Detect faces

        :param image_file: source image file path
        :param image_uri: Google Cloud Storage URI
        :param json_file: json target to save result
        :return: detection response

        **Examples**

        **Robot Framework**

        .. code-block:: robotframework

            ${result}=   Face Detection   image_uri=gs://vision/faces.png
            ...  json_file=${CURDIR}${/}result.json
        """
        parameters = self.set_image_type(image_file, image_uri)
        response = self.service.face_detection(**parameters)
        self.write_json(json_file, response)
        return response



from typing import Dict, Union, Optional

from google.cloud import translate_v3

from . import LibraryContext, keyword, TextType, to_texttype


class TranslationKeywords(LibraryContext):
    """Class for Google Cloud Translation API

    Link to `Translation PyPI`_ page.

    .. _Translation PyPI: https://pypi.org/project/google-cloud-translate/
    """

    def __init__(self, ctx):
        super().__init__(ctx)
        self.service = None
        self.project_id = None

    @keyword(tags=["init", "translation"])
    def init_translation(
        self,
        project_identifier: str,
        service_account: str = None,
        use_robocorp_vault: Optional[bool] = None,
        token_file: str = None,
    ) -> None:
        """Initialize Google Cloud Translation client

        :param project_identifier: identifier for Translation project
        :param service_account: file path to service account file
        :param use_robocorp_vault: use credentials in `Robocorp Vault`
        :param token_file: file path to token file
        """
        self.project_id = project_identifier
        self.service = self.init_service_with_object(
            translate_v3.TranslationServiceClient,
            service_account,
            use_robocorp_vault,
            token_file,
        )

    @keyword(tags=["translation"])
    def translate(
        self,
        text: Union[list, str],
        source_language: str = None,
        target_language: str = None,
        mime_type: TextType = None,
    ) -> Dict:
        """Translate text

        :param text: text to translate
        :param source_language: language code
        :param target_language: language code
        :param mime_type: text or html
        :return: translated text

        **Examples**

        **Robot Framework**

        .. code-block:: robotframework

            ${result}=   Translate   ${TEXT}  target_language=de
        """
        if not text and not target_language:
            raise KeyError("text and target_language are required parameters")
        parent = f"projects/{self.project_id}"
        contents = [text] if not isinstance(text, list) else text

        parameters = {"parent": parent, "contents": contents}
        if source_language:
            parameters["source_language_code"] = source_language
        if target_language:
            parameters["target_language_code"] = target_language
        if mime_type:
            mimetype = to_texttype(mime_type)
            parameters["mime_type"] = mimetype
        response = self.service.translate_text(**parameters)
        return response



# flake8: noqa

from robot.api.deco import keyword
from .context import (
    LibraryContext,
    ElementNotFound,
    MultipleElementsFound,
    TimeoutException,
    GoogleOAuthAuthenticationError,
)

from .enums import TextType, to_texttype, UpdateAction, VideoFeature, to_feature
from .base import BaseKeywords

from .apps_script import AppsScriptKeywords
from .drive import DriveKeywords
from .gmail import GmailKeywords
from .natural_language import NaturalLanguageKeywords
from .sheets import SheetsKeywords
from .speech_to_text import SpeechToTextKeywords
from .storage import StorageKeywords
from .text_to_speech import TextToSpeechKeywords
from .translation import TranslationKeywords
from .video_intelligence import VideoIntelligenceKeywords
from .vision import VisionKeywords



from io import BytesIO
import mimetypes
from pathlib import Path
import shutil
import time
from typing import Dict, List, Optional

from apiclient.errors import HttpError
from apiclient.http import MediaFileUpload, MediaIoBaseDownload

from . import LibraryContext, keyword, UpdateAction


class GoogleDriveError(Exception):
    """Raised with errors in Drive API"""


class DriveKeywords(LibraryContext):
    """Class for Google Drive API

    For more information about Google Drive API link_.

    .. _link: https://developers.google.com/drive/api
    """

    def __init__(self, ctx):
        super().__init__(ctx)
        self.service = None

    @keyword(tags=["init", "drive"])
    def init_drive(
        self,
        service_account: str = None,
        credentials: str = None,
        use_robocorp_vault: Optional[bool] = None,
        scopes: list = None,
        token_file: str = None,
    ) -> None:
        """Initialize Google Drive client

        :param service_account: file path to service account file
        :param credentials: file path to credentials file
        :param use_robocorp_vault: use credentials in `Robocorp Vault`
        :param scopes: list of extra authentication scopes
        :param token_file: file path to token file
        """
        drive_scopes = [
            "drive",
            "drive.appdata",
            "drive.file",
            "drive.install",
            "drive.metadata",
        ]
        if scopes:
            drive_scopes += scopes
        self.service = self.init_service(
            service_name="drive",
            api_version="v3",
            scopes=drive_scopes,
            service_account_file=service_account,
            credentials_file=credentials,
            use_robocorp_vault=use_robocorp_vault,
            token_file=token_file,
        )

    @keyword(tags=["drive"])
    def upload_drive_file(
        self,
        filename: str = None,
        folder: str = None,
        overwrite: bool = False,
        make_dir: bool = False,
    ) -> str:
        """Upload files into Drive

        :param filename: name of the file to upload
        :param folder: target folder for upload
        :param overwrite: set to `True` if already existing file should be overwritten
        :param make_dir: set to `True` if folder should be created if it does not exist
        :return: uploaded file id

        Example:

        .. code-block:: robotframework

            ${file1_id}=  Upload Drive File  data.json  # Upload file to drive root
            ${file2_id}=  Upload Drive File  newdata.json  new_folder  make_dir=True
            ${file3_id}=  Upload Drive File  data.json  overwrite=True
        """
        folder_id = self.get_drive_folder_id(folder)
        if folder_id is None and make_dir:
            folder_id = self.create_drive_directory(folder)
        if folder_id is None:
            raise GoogleDriveError(
                "Target folder '%s' does not exist or could not be created" % folder
            )

        filepath = Path(filename)
        if filepath.is_dir():
            raise GoogleDriveError(
                "The '%s' is a directory and can not be uploaded" % filename
            )
        elif not filepath.is_file():
            raise GoogleDriveError("Filename '%s' does not exist" % filename)

        query_string = f"name = '{filepath.name}' and '{folder_id}' in parents"
        target_file = self.search_drive_files(query=query_string, recurse=True)
        guess_mimetype = mimetypes.guess_type(str(filepath.absolute()))
        file_mimetype = guess_mimetype[0] if guess_mimetype else "*/*"
        media = MediaFileUpload(
            filepath.absolute(), mimetype=file_mimetype, resumable=True
        )
        file_metadata = {
            "name": filepath.name,
            "parents": [folder_id],
            "mimeType": file_mimetype,
        }
        if len(target_file) == 1 and overwrite:
            self.logger.info("Overwriting file '%s' with new content", filename)
            file = (
                self.service.files()
                .update(fileId=target_file[0]["id"], media_body=media, fields="id")
                .execute()
            )
            return file.get("id", None)
        elif len(target_file) == 1 and not overwrite:
            self.logger.warn("Not uploading new copy of file '%s'", filepath.name)
            return None
        elif len(target_file) > 1:
            self.logger.warn(
                "Drive already contains '%s' copies of file '%s'. Not uploading again."
                % (len(target_file), filepath.name)
            )
            return None
        else:
            file = (
                self.service.files()
                .create(
                    body=file_metadata,
                    media_body=media,
                    fields="id",
                )
                .execute()
            )
            return file.get("id", None)

    def _download_with_fileobject(self, file_object):
        request = self.service.files().get_media(fileId=file_object["id"])
        fh = BytesIO()
        downloader = MediaIoBaseDownload(fh, request)
        done = False
        while done is False:
            _, done = downloader.next_chunk()
        fh.seek(0)
        with open(file_object["name"], "wb") as f:
            # pylint: disable=protected-access
            shutil.copyfileobj(fh, f, length=downloader._total_size)

    @keyword(tags=["drive"])
    def download_drive_files(
        self,
        file_dict: dict = None,
        query: str = None,
        source: str = None,
        limit: int = None,
        timeout: float = None,
    ) -> List:
        """Download files specified by file dictionary or query string

        Parameters `start`, `limit` and `timeout` are used only when
        downloading files defined by `query` parameter.

        :param file_dict: file dictionary returned by `Search Drive Files`
        :param query: drive query string to find target files, defaults to None
        :param source: source directory where query is executed
        :param limit: maximum amount of files that are downloaded, defaults to None
        :param timeout: maximum allowed time in seconds for download process
        :return: list of downloaded files

        Example:

        .. code-block:: robotframework

            ${files}=    Search Drive Files    query=name contains '.json'
            FOR    ${f}    IN    @{files}
                IF  ${f}[size] < 2000
                    Download Drive Files  file_dict=${f}
                END
            END

            ${folder_id}=   Get Drive Folder Id   datafolder
            Download Drive Files  query=name contains '.json' and '${folder_id}' in parents  recurse=True
        """  # noqa: E501
        if query:
            filelist = self.search_drive_files(query, source=source)
            files_downloaded = []
            start_time = time.time()

            for f in filelist:
                self._download_with_fileobject(f)
                current_time = time.time()
                files_downloaded.append(f["name"])
                if limit and len(files_downloaded) >= limit:
                    self.logger.info(
                        "Drive download limit %s reached. Stopping the download.", limit
                    )
                    break
                if timeout and (current_time - start_time) > float(timeout):
                    self.logger.info(
                        "Drive download timeout %s seconds reached. "
                        "Stopping the download.",
                        timeout,
                    )
                    break
            return files_downloaded

        if file_dict:
            self._download_with_fileobject(file_dict)
            return [file_dict]
        return []

    @keyword(tags=["drive"])
    def update_drive_file(
        self,
        file_id: str = None,
        file_dict: dict = None,
        query: str = None,
        source: str = None,
        action: UpdateAction = UpdateAction.star,
        multiple_ok: bool = False,
    ) -> int:
        """Update file specified by id, file dictionary or query string

        Possible actions:
        - star
        - unstar
        - trash
        - untrash

        :param file_id: drive file id
        :param file_dict: file dictionary returned by `Drive Search Files`
        :param query: drive query string to find target file, needs to match 1 file
        :param source: source directory where query is executed
        :param action: update action, default star file
        :param multiple_ok: set to `True` if it is ok to perform update
         on more than 1 file
        :return: number of updated files

        Example:

        .. code-block:: robotframework

            ${folder_id}=  Get Drive Folder Id   datafolder
            ${updated}=    Update Drive File  query=name contains '.json' and '${folder_id}' in parents
            ...            action=star
            ...            multiple_ok=True
        """  # noqa: E501
        target_files = self._get_target_file(
            file_id, file_dict, query, multiple_ok, source
        )
        update_count = 0
        for tf in target_files:
            self._drive_files_update(tf, action)
            update_count += 1
        return update_count

    def _get_target_file(self, file_id, file_dict, query, multiple_ok, source=None):
        target_files = []
        if file_id:
            target_files.append(file_id)
        elif file_dict:
            target_files.append(file_dict.get("id", None))
        else:
            files = self.search_drive_files(query, source=source, recurse=True)
            target_files = [tf.get("id", None) for tf in files]
            if not multiple_ok and len(target_files) > 1:
                raise GoogleDriveError(
                    "expected search to match 1 file, but it matched %s files"
                    % len(files)
                )

        return target_files

    def _drive_files_update(self, file_id: str, action: UpdateAction):
        body = None
        if action == UpdateAction.trash:
            body = {"trashed": True}
        elif action == UpdateAction.untrash:
            body = {"trashed": False}
        elif action == UpdateAction.star:
            body = {"starred": True}
        elif action == UpdateAction.unstar:
            body = {"starred": False}
        else:
            # TODO: mypy should handle enum exhaustivity validation
            raise ValueError(f"Unsupported update action: {action}")
        self.logger.info(body)
        updated_file = self.service.files().update(fileId=file_id, body=body).execute()
        return updated_file

    @keyword(tags=["drive"])
    def delete_drive_file(
        self,
        file_id: str = None,
        file_dict: dict = None,
        query: str = None,
        multiple_ok: bool = False,
    ) -> int:
        """Delete file specified by id, file dictionary or query string

        Note. Be extra careful when calling this keyword!

        :param file_id: drive file id
        :param file_dict: file dictionary returned by `Search Drive Files`
        :param query: drive query string to find target file, needs to match 1 file
         unless parameter `multiple_ok` is set to `True`
        :param multiple_ok: set to `True` if it is ok to perform delete
         on more than 1 file
        :return: how many files where deleted

        Example:

        .. code-block:: robotframework

            ${folder_id}=  Get Drive Folder Id   datafolder
            ${deleted}=    Delete Drive File  query=name contains '.json' and '${folder_id}' in parents
            ...            multiple_ok=True
        """  # noqa: E501
        target_files = self._get_target_file(file_id, file_dict, query, multiple_ok)

        delete_count = 0
        for tf in target_files:
            self.service.files().delete(fileId=tf).execute()
            delete_count += 1
        return delete_count

    @keyword(tags=["drive"])
    def get_drive_folder_id(self, folder: str = None) -> str:
        """Get file id for the folder

        :param folder: name of the folder to identify, by default returns drive's
         `root` folder id
        :return: file id of the folder

        Example:

        .. code-block:: robotframework

            ${root_id}=    Get Drive Folder Id   # returns Drive root folder id
            ${folder_id}=  Get Drive Folder Id  subdir
        """
        mime_folder_type = "application/vnd.google-apps.folder"
        folder_id = None
        if folder is None:
            file = self.service.files().get(fileId="root", fields="id").execute()
            folder_id = file.get("id", None)
        else:
            query_string = f"name = '{folder}' AND mimeType = '{mime_folder_type}'"
            folders = self.search_drive_files(query=query_string, recurse=True)
            if len(folders) == 1:
                folder_id = folders[0].get("id", None)
            else:
                self.logger.info(
                    "Found %s directories with name '%s'" % (len(folders), folder)
                )
        return folder_id

    @keyword(tags=["drive"])
    def move_drive_file(
        self,
        file_id: str = None,
        file_dict: dict = None,
        query: str = None,
        source: str = None,
        target: str = None,
        multiple_ok: bool = False,
    ) -> List:
        """Move file specified by id, file dictionary or query string into target folder

        :param file_id: drive file id
        :param file_dict: file dictionary returned by `Search Drive Files`
        :param query: drive query string to find target file, needs to match 1 file
        :param source: name of the folder to move file from, is by default drive's
         `root` folder id
        :param target: name of the folder to move file into, is by default drive's
         `root` folder id
        :param multiple_ok: if `True` then moving more than 1 file
        :return: list of file ids

        Example:

        .. code-block:: robotframework

            ${source_id}=  Get Drive Folder Id  sourcefolder
            ${query}=      Set Variable  name contains '.json' and '${sourceid}' in parents
            ${files}=      Move Drive File  query=${query}  folder=target_folder  multiple_ok=True
        """  # noqa: E501
        result_files = []
        if file_id or file_dict:
            target_files = self._get_target_file(file_id, file_dict, query, multiple_ok)
        else:
            target_files = self.search_drive_files(query, source=source)
        target_parent = None
        if len(target_files) == 0:
            raise GoogleDriveError("Did not find any files to move")
        if target:
            target_parent = self.get_drive_folder_id(target)
        if target_parent is None:
            raise GoogleDriveError(
                "Unable to find target folder: '%s'" % (target if target else "root")
            )
        for tf in target_files:
            file = self.service.files().get(fileId=tf["id"], fields="parents").execute()
            previous_parents = ",".join(file.get("parents"))
            result_file = (
                self.service.files()
                .update(
                    fileId=tf["id"],
                    addParents=target_parent,
                    removeParents=previous_parents,
                    fields="id, parents",
                )
                .execute()
            )
            result_files.append(result_file)
        return result_files

    @keyword(tags=["drive"])
    def search_drive_files(
        self, query: str = None, recurse: bool = False, source: str = None
    ) -> List:
        """Search Google Drive for files matching query string

        :param query: search string, defaults to None which means that all files
         and folders are returned
        :param recurse: set to `True` if search should recursive
        :param source: source directory where query is executed
        :return: list of files

        Example:

        .. code-block:: robotframework

            ${files}=  Search Drive Files   query=name contains 'hello'
            ${files}=  Search Drive Files   query=modifiedTime > '2020-06-04T12:00:00'
            ${files}=  Search Drive Files   query=mimeType contains 'image/' or mimeType contains 'video/'
            ${files}=  Search Drive Files   query=name contains '.yaml'  recurse=True
            ${files}=  Search Drive Files   query=name contains '.yaml'  source=datadirectory
        """  # noqa: E501
        page_token = None
        filelist = []
        parameters = {
            "fields": "nextPageToken, files(id, name, mimeType, parents)",
            "q": query,
        }

        if not recurse:
            folder_id = "root" if not source else self.get_drive_folder_id(source)
            if folder_id is None:
                return []
            parameters["q"] += f" and '{folder_id}' in parents"

        while True:
            if page_token:
                parameters["pageToken"] = page_token
            try:
                response = self.service.files().list(**parameters).execute()
                self.logger.info(response)
                for file in response.get("files", []):
                    filesize = file.get("size")
                    filelist.append(
                        {
                            "name": file.get("name"),
                            "id": file.get("id"),
                            "size": int(filesize) if filesize else None,
                            "kind": file.get("kind"),
                            "parents": file.get("parents"),
                            "starred": file.get("starred"),
                            "trashed": file.get("trashed"),
                            "shared": file.get("shared"),
                            "mimeType": file.get("mimeType"),
                            "spaces": file.get("spaces", None),
                            "exportLinks": file.get("exportLinks"),
                            "createdTime": file.get("createdTime"),
                            "modifiedTime": file.get("modifiedTime"),
                        }
                    )
                page_token = response.get("nextPageToken", None)
                if page_token is None:
                    break
            except HttpError as e:
                raise GoogleDriveError(str(e)) from e
        return filelist

    @keyword(tags=["drive"])
    def create_drive_directory(
        self, folder: str = None, parent_folder: str = None
    ) -> Dict:
        """Create new directory to Google Drive

        :param folder: name for the new directory
        :param parent_folder: top level directory for new directory
        :return: created file id
        """
        if not folder or len(folder) == 0:
            raise GoogleDriveError("Can't create Drive directory with empty name")

        folder_id = self.get_drive_folder_id(folder)  # , parent_folder)
        if folder_id:
            self.logger.info(
                "Folder '%s' already exists. Not creating new one.", folder_id
            )
            return None

        file_metadata = {
            "name": folder,
            "mimeType": "application/vnd.google-apps.folder",
        }

        if parent_folder:
            parent_folder_id = self.get_drive_folder_id(parent_folder)
            file_metadata["parents"] = [parent_folder_id]
        added_folder = (
            self.service.files().create(body=file_metadata, fields="id").execute()
        )
        return added_folder

    @keyword(tags=["drive"])
    def export_drive_file(
        self,
        file_id: str = None,
        file_dict: dict = None,
        target_file: str = None,
        mimetype: str = "application/pdf",
    ) -> str:
        """Export Google Drive file using Drive export links

        :param file_id: drive file id
        :param file_dict: file dictionary returned by `Search Drive Files`
        :param target_file: name for the exported file
        :param mimetype: export mimetype, defaults to "application/pdf"
        :return: file path to the exported file

        Example:

        .. code-block:: robotframework

            ${files}=  Drive Search Files  query=name contains 'my example worksheet'
            Export Drive File  file_dict=${files}[0]
        """
        if target_file is None or len(target_file) == 0:
            raise AttributeError("The target_file is required parameter for export")
        if file_id is None and file_dict is None:
            raise AttributeError("Either file_id or file_dict is required for export")
        target_files = self._get_target_file(file_id, file_dict, None, False)
        if len(target_files) != 1:
            raise ValueError("Did not find the Google Drive file to export")
        request = self.service.files().export(fileId=target_files[0], mimeType=mimetype)

        fh = BytesIO()
        downloader = MediaIoBaseDownload(fh, request)
        done = False
        while done is False:
            _, done = downloader.next_chunk()

        fh.seek(0)
        filepath = Path(target_file).absolute()
        with open(filepath, "wb") as f:
            shutil.copyfileobj(fh, f, length=131072)
        return filepath



from . import (
    LibraryContext,
    keyword,
)


class BaseKeywords(LibraryContext):
    """Base keywords for the Google library"""

    @keyword
    def set_robocorp_vault(
        self,
        vault_name: str = None,
        vault_secret_key: str = None,
        auth_type: str = "serviceaccount",
    ):
        """Set Robocorp Vault name and secret key name
        :param vault_name: Robocorp Vault name
        :param vault_secret_key: Robocorp Vault secret key name
        """
        if vault_name:
            self.ctx.robocorp_vault_name = vault_name
        if vault_secret_key:
            self.ctx.robocorp_vault_secret_key = vault_secret_key
        if self.ctx.robocorp_vault_name and self.ctx.robocorp_vault_secret_key:
            self.ctx.use_robocorp_vault = True
        self.ctx.cloud_auth_type = auth_type



import importlib
import logging
import os
from robotlibcore import DynamicCore


from .keywords import (
    AppsScriptKeywords,
    BaseKeywords,
    DriveKeywords,
    GmailKeywords,
    NaturalLanguageKeywords,
    SheetsKeywords,
    SpeechToTextKeywords,
    StorageKeywords,
    TextToSpeechKeywords,
    TranslationKeywords,
    VideoIntelligenceKeywords,
    VisionKeywords,
)


def import_vault():
    """Try to import Vault/Secrets library, with old and new name."""
    try:
        module = importlib.import_module("RPA.Robocorp.Vault")
        return getattr(module, "Vault")
    except ModuleNotFoundError:
        pass

    try:
        module = importlib.import_module("RPA.Robocloud.Secrets")
        return getattr(module, "Secrets")
    except ModuleNotFoundError:
        pass

    return None


class Google(DynamicCore):
    """`Google` is a library for operating with Google API endpoints.

    Usage requires the following steps:

    - Create a GCP project
    - Enable approriate APIs
    - Create credentials (OAuth or service account)
    - Install rpaframework-google package

    **Google authentication**

    Authentication for Google is set with `service account JSON file` which can be given to the library
    in three different ways or with `OAuth2 token`, which is used for OAuth authentication.

    Methods when using service account:

    - Method 1 as environment variables, ``GOOGLE_APPLICATION_CREDENTIALS`` with path to service account file.
    - Method 2 as keyword parameter to ``Init Storage`` for example.
    - Method 3 as Robocorp vault secret. The vault name and secret key name needs to be given in library init
      or with keyword ``Set Robocorp Vault``. Secret value should contain JSON file contents.

    Method 1. service account using environment variable

    .. code-block:: robotframework

        *** Settings ***
        Library   RPA.Cloud.Google

        *** Tasks ***
        Init Google services
            # NO parameters for Init Vision, expecting to get JSON
            # with GOOGLE_APPLICATION_CREDENTIALS environment variable
            Init Vision

    Method 2. service account with keyword parameter

    .. code-block:: robotframework

        *** Settings ***
        Library   RPA.Cloud.Google

        *** Tasks ***
        Init Google services
            Init Speech To Text   /path/to/service_account.json

    Method 3. setting Robocorp Vault in the library init

    .. code-block:: robotframework

        *** Settings ***
        Library   RPA.Cloud.Google
        ...       vault_name=googlecloud
        ...       vault_secret_key=servicecreds

        *** Tasks ***
        Init Google services
            Init Storage

    Method 3. setting Robocorp Vault with keyword

    .. code-block:: robotframework

        *** Settings ***
        Library   RPA.Cloud.Google

        *** Tasks ***
        Init Google services
            Set Robocorp Vault   vault_name=googlecloud  vault_secret_key=servicecreds
            Init Storage    use_robocorp_vault=${TRUE}

    Methods when using OAuth token:

    - Method 1 as keyword parameter ``token_file`` to ``Init Storage`` for example.
    - Method 2 as Robocorp vault secret. The vault name and secret key name needs to be given in library init
      or with keyword ``Set Robocorp Vault``. Secret value should contain JSON file contents.

    Method 1. The Google Apps Script and Google Drive services are authenticated using this method.

    .. code-block:: robotframework

        *** Settings ***
        Library   RPA.Cloud.Google

        *** Variables ***
        @{SCRIPT_SCOPES}     forms   spreadsheets

        *** Tasks ***
        Init Google OAuth services
            Init Apps Script    token_file=oauth_token   ${SCRIPT_SCOPES}

    Method 2. setting Robocorp Vault in the library init

    .. code-block:: robotframework

        *** Settings ***
        Library   RPA.Cloud.Google
        ...       vault_name=googlecloud
        ...       vault_secret_key=oauth
        ...       cloud_auth_type=token

        *** Tasks ***
        Init Google services
            Init Storage


    **Creating and using OAuth token file**

    The token file can be created using `credentials.json` by running command:

    ``rpa-google-oauth --credentials <filepath> --service drive`` or
    ``rpa-google-oauth --credentials <filepath> --scopes drive.appdata,drive.file,drive.install``

    This will start web based authentication process, which outputs the token at the end.
    Token could be stored into ``Robocorp Vault``.

    Example Vault content.

    .. code-block:: json

        "googlecloud": {
            "oauth-token": "gANfd123321aabeedYsc"
        }

    Using the Vault.

    .. code-block:: robotframework

        *** Keywords ***
        Set up Google Drive authentication
            Set Robocorp Vault   vault_name=googlecloud
            ...  vault_secret_key=oauth-token
            ...  cloud_auth_type=token
            Init Drive


    **Installation**

    ``pip install rpaframework-google``

    **Examples**

    **Robot Framework**

    .. code-block:: robotframework

        *** Settings ***
        Library   RPA.Cloud.Google

        *** Variables ***
        ${SERVICE_ACCOUNT}    ${/}path${/}to${/}service_account.json
        ${BUCKET_NAME}            testbucket12213123123

        *** Tasks ***
        Upload a file into a new storage bucket
            [Setup]   Init Storage    ${SERVICE_ACCOUNT}
            Create Storage Bucket    ${BUCKET_NAME}
            Upload Storage File      ${BUCKET_NAME}
            ...   ${/}path${/}to${/}file.pdf
            ...   myfile.pdf
            @{files}         List Storage Files   ${BUCKET_NAME}
            FOR   ${file}  IN   @{files}
                Log  ${file}
            END

    **Python**

    .. code-block:: python

        from RPA.Cloud.Google import Google

        library = Google
        service_account = '/path/to/service_account.json'

        library.init_vision(service_account)
        library.init_text_to_speech(service_account)

        response = library.detect_text('imagefile.png', 'result.json')
        library.synthesize_speech('I want this said aloud', target_file='said.mp3')
    """  # noqa: E501

    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    ROBOT_LIBRARY_DOC_FORMAT = "REST"

    def __init__(
        self,
        service_account: str = None,
        vault_name: str = None,
        vault_secret_key: str = None,
        cloud_auth_type: str = "serviceaccount",
    ):
        """Library initialization

        :param service_account: path to service account
        :param vault_name: Robocorp vault name
        :param vault_secret_key: Robocorp secret key
        :param cloud_auth_type: "serviceaccount" or "token",
         defaults to "serviceaccount"
        """
        self.logger = logging.getLogger(__name__)
        self.service_account_file = service_account
        self.robocorp_vault_name = vault_name
        self.robocorp_vault_secret_key = vault_secret_key
        self.cloud_auth_type = cloud_auth_type
        self.use_robocorp_vault = False
        if self.robocorp_vault_name and self.robocorp_vault_secret_key:
            self.use_robocorp_vault = True
        if self.service_account_file is None:
            self.service_account_file = os.getenv("GOOGLE_APPLICATION_CREDENTIALS")
        self.secrets_library = import_vault()

        # Register keyword libraries to LibCore
        libraries = [
            AppsScriptKeywords(self),
            BaseKeywords(self),
            DriveKeywords(self),
            GmailKeywords(self),
            NaturalLanguageKeywords(self),
            SheetsKeywords(self),
            SpeechToTextKeywords(self),
            StorageKeywords(self),
            TextToSpeechKeywords(self),
            TranslationKeywords(self),
            VideoIntelligenceKeywords(self),
            VisionKeywords(self),
        ]
        super().__init__(libraries)



from typing import Dict, List, Optional

from . import (
    LibraryContext,
    keyword,
)


class SheetsKeywords(LibraryContext):
    """Keywords for Google Sheets operations"""

    def __init__(self, ctx):
        super().__init__(ctx)
        self.service = None

    @keyword(tags=["init", "sheets"])
    def init_sheets(
        self,
        service_account: str = None,
        credentials: str = None,
        use_robocorp_vault: Optional[bool] = None,
        scopes: list = None,
        token_file: str = None,
    ) -> None:
        """Initialize Google Sheets client

        :param service_account: file path to service account file
        :param credentials: file path to credentials file
        :param use_robocorp_vault: use credentials in `Robocorp Vault`
        :param scopes: list of extra authentication scopes
        :param token_file: file path to token file
        """
        sheets_scopes = ["drive", "drive.file", "spreadsheets"]
        if scopes:
            sheets_scopes += scopes
        self.service = self.init_service(
            service_name="sheets",
            api_version="v4",
            scopes=sheets_scopes,
            service_account_file=service_account,
            credentials_file=credentials,
            use_robocorp_vault=use_robocorp_vault,
            token_file=token_file,
        )

    @keyword(tags=["sheets"])
    def create_sheet(self, title: str) -> str:
        """Create empty sheet with a title

        :param title: name as string
        :return: created `sheet_id`

        **Examples**

        **Robot Framework**

        .. code-block:: robotframework

            ${result}=  Create Sheet   Example Sheet
        """
        if not title:
            raise KeyError("title is required for kw: create_sheet")

        data = {"properties": {"title": title}}
        spreadsheet = (
            self.service.spreadsheets()
            .create(body=data, fields="spreadsheetId")
            .execute()
        )
        return spreadsheet.get("spreadsheetId")

    @keyword(tags=["sheets"])
    def insert_sheet_values(
        self,
        sheet_id: str,
        sheet_range: str,
        values: list,
        major_dimension: str = "COLUMNS",
        value_input_option: str = "USER_ENTERED",
    ) -> Dict:
        """Insert values into sheet cells

        :param sheet_id: target sheet
        :param sheet_range: target sheet range
        :param values: list of values to insert into sheet
        :param major_dimension: major dimension of the values, default `COLUMNS`
        :param value_input_option: controls whether input strings are parsed or not,
         default `USER_ENTERED`
        :return: operation result

        **Examples**

        **Robot Framework**

        .. code-block:: robotframework

            ${values}   Evaluate   [[11, 12, 13], ['aa', 'bb', 'cc']]
            ${result}=  Insert Sheet Values   ${SHEET_ID}  A:B  ${values}
            ${result}=  Insert Sheet Values   ${SHEET_ID}  A:B  ${values}  ROWS
        """
        resource = {"majorDimension": major_dimension, "values": values}
        return (
            self.service.spreadsheets()
            .values()
            .append(
                spreadsheetId=sheet_id,
                range=sheet_range,
                body=resource,
                valueInputOption=value_input_option,
            )
            .execute()
        )

    @keyword(tags=["sheets"])
    def update_sheet_values(
        self,
        sheet_id: str,
        sheet_range: str,
        values: list,
        major_dimension: str = "COLUMNS",
        value_input_option: str = "USER_ENTERED",
    ) -> Dict:
        """Insert values into sheet cells

        :param sheet_id: target sheet
        :param sheet_range: target sheet range
        :param values: list of values to insert into sheet
        :param major_dimension: major dimension of the values, default `COLUMNS`
        :param value_input_option: controls whether input strings are parsed or not,
         default `USER_ENTERED`
        :return: operation result

        **Examples**

        **Robot Framework**

        .. code-block:: robotframework

            ${row}  Evaluate   [[22, 33 ,44]]
            ${result}=  Update Sheet Values  ${SHEET_ID}  A6:C6  ${row}   ROWS
        """
        resource = {"majorDimension": major_dimension, "values": values}
        return (
            self.service.spreadsheets()
            .values()
            .update(
                spreadsheetId=sheet_id,
                range=sheet_range,
                body=resource,
                valueInputOption=value_input_option,
            )
            .execute()
        )

    @keyword(tags=["sheets"])
    def get_sheet_values(
        self,
        sheet_id: str,
        sheet_range: str,
        value_render_option: str = "UNFORMATTED_VALUE",
        datetime_render_option: str = "FORMATTED_STRING",
    ) -> List:
        """Get values from the range in the sheet

        :param sheet_id: target sheet
        :param sheet_range: target sheet range
        :param value_render_option: how values should be represented
         in the output defaults to "UNFORMATTED_VALUE"
        :param datetime_render_option: how dates, times, and durations should be
         represented in the output, defaults to "FORMATTED_STRING"
        :return: operation result

        **Examples**

        **Robot Framework**

        .. code-block:: robotframework

            ${values}=  Get Sheet Values  ${SHEET_ID}  A1:C1
        """  # noqa: E501
        return (
            self.service.spreadsheets()
            .values()
            .get(
                spreadsheetId=sheet_id,
                range=sheet_range,
                valueRenderOption=value_render_option,
                dateTimeRenderOption=datetime_render_option,
            )
            .execute()
        )

    @keyword(tags=["sheets"])
    def clear_sheet_values(self, sheet_id: str, sheet_range: str) -> Dict:
        """Clear cell values for range of cells within a sheet

        :param sheet_id: target sheet
        :param sheet_range: target sheet range
        :return: operation result

        **Examples**

        **Robot Framework**

        .. code-block:: robotframework

            ${result}=  Clear Sheet Values  ${SHEET_ID}  A1:C1
        """
        return (
            self.service.spreadsheets()
            .values()
            .clear(
                spreadsheetId=sheet_id,
                range=sheet_range,
            )
            .execute()
        )

    @keyword(tags=["sheets"])
    def copy_sheet(self, sheet_id: str, target_sheet_id: str) -> Dict:
        """Copy spreadsheet to target spreadsheet

        *NOTE:* service account user must have access to
        target sheet also

        :param sheet_id: ID of the sheet to copy
        :param target_sheet_id: ID of the target sheet
        :return: operation result

        **Examples**

        **Robot Framework**

        .. code-block:: robotframework

            ${result}=  Copy Sheet   ${SHEET_ID}  ${NEW_SHEET}
        """
        body = {
            "destination_spreadsheet_id": target_sheet_id,
        }
        return (
            self.service.spreadsheets()
            .sheets()
            .copyTo(
                spreadsheetId=sheet_id,
                sheetId=0,
                body=body,
            )
            .execute()
        )
